import ubelt as ub
import pytest
from os.path import basename, join, exists


@pytest.mark.timeout(5)
def test_download_no_fpath():
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.ensure_app_cache_dir('ubelt')
    fname = basename(url)
    fpath = join(dpath, fname)

    ub.delete(fpath)
    assert not exists(fpath)

    got_fpath = ub.download(url)

    assert got_fpath == fpath
    assert exists(fpath)


@pytest.mark.timeout(5)
def test_download_with_fpath():
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url(1201)

    dpath = ub.ensure_app_cache_dir('ubelt', 'tests')
    fname = basename(url)
    fpath = join(dpath, fname)

    ub.delete(fpath)
    assert not exists(fpath)

    got_fpath = ub.download(url, fpath=fpath)
    assert got_fpath == fpath
    assert exists(fpath)

    with open(got_fpath, 'rb') as file:
        data = file.read()
    assert len(data) > 1200, 'should have downloaded some bytes'


@pytest.mark.timeout(5)
def test_download_chunksize():
    # url = 'https://www.dropbox.com/s/jl506apezj42zjz/ibeis-win32-setup-ymd_hm-2015-08-01_16-28.exe?dl=1'
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.ensure_app_cache_dir('ubelt')
    fname = basename(url)
    fpath = join(dpath, fname)

    ub.delete(fpath)
    assert not exists(fpath)

    got_fpath = ub.download(url, chunksize=2)

    assert got_fpath == fpath
    assert exists(fpath)


@pytest.mark.timeout(5)
def test_download_cover_hashers():
    # url = 'https://www.dropbox.com/s/jl506apezj42zjz/ibeis-win32-setup-ymd_hm-2015-08-01_16-28.exe?dl=1'
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.ensure_app_cache_dir('ubelt')
    fname = basename(url)

    # add coverage for different hashers
    ub.download(url, hasher='md5', hash_prefix='e09c80c42fda55f9d992e59ca6b33',
                dpath=dpath, fname=fname)
    ub.download(url, hasher='sha256', hash_prefix='bf2cb58a68f684d95a3b78ef8f',
                dpath=dpath, fname=fname)


@pytest.mark.timeout(5)
def test_download_hashalgo():
    # url = 'https://www.dropbox.com/s/jl506apezj42zjz/ibeis-win32-setup-ymd_hm-2015-08-01_16-28.exe?dl=1'
    import hashlib

    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.ensure_app_cache_dir('ubelt')
    fname = basename(url)
    fpath = join(dpath, fname)

    ub.delete(fpath)
    assert not exists(fpath)

    got_fpath = ub.download(url,
                            hash_prefix='e09c80c42fda55f9d992e59ca6b3307d',
                            hasher=hashlib.md5())

    assert got_fpath == fpath
    assert exists(fpath)


@pytest.mark.timeout(5)
def test_grabdata_cache():
    """
    Check where the url is downloaded to when fpath is not specified.
    """
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.ensure_app_cache_dir('ubelt')
    fname = basename(url)
    fpath = join(dpath, fname)

    got_fpath = ub.grabdata(url)
    assert got_fpath == fpath
    assert exists(fpath)

    ub.delete(fpath)
    assert not exists(fpath)

    ub.grabdata(url)
    assert exists(fpath)


@pytest.mark.timeout(5)
def test_grabdata_url_only():
    """
    Check where the url is downloaded to when fpath is not specified.
    """
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.ensure_app_cache_dir('ubelt')
    fname = basename(url)
    fpath = join(dpath, fname)

    got_fpath = ub.grabdata(url)
    assert got_fpath == fpath
    assert exists(fpath)


@pytest.mark.timeout(5)
def test_grabdata_with_fpath():
    """
    Check where the url is downloaded to when fpath is not specified.
    """
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.ensure_app_cache_dir('ubelt')
    fname = basename(url)
    fpath = join(dpath, fname)

    got_fpath = ub.grabdata(url, fpath=fpath, verbose=3)
    assert got_fpath == fpath
    assert exists(fpath)

    ub.delete(fpath)
    assert not exists(fpath)

    ub.grabdata(url, fpath=fpath, verbose=3)
    assert exists(fpath)


def test_grabdata_value_error():
    """
    Check where the url is downloaded to when fpath is not specified.
    """
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.ensure_app_cache_dir('ubelt')
    fname = basename(url)
    fpath = join(dpath, fname)

    with pytest.raises(ValueError):
        ub.grabdata(url, fname=fname, fpath=fpath, dpath=dpath)

    with pytest.raises(ValueError):
        ub.grabdata(url, fname=fname, fpath=fpath)

    with pytest.raises(ValueError):
        ub.grabdata(url, dpath=dpath, fpath=fpath)

    with pytest.raises(ValueError):
        ub.grabdata(url, fpath=fpath, appname='foobar')

    with pytest.raises(ValueError):
        ub.grabdata(url, dpath=dpath, appname='foobar')


@pytest.mark.timeout(10)
def test_download_bad_url():
    """
    Check that we error when the url is bad

    Notes:
        For some reason this can take a long time to realize there is no URL,
        even if the timeout is specified and fairly low.

    CommandLine:
        python tests/test_download.py test_download_bad_url --verbose
    """
    import pytest
    pytest.skip('This takes a long time to timeout and I dont understand why')

    url = 'http://www.a-very-incorrect-url.gov/does_not_exist.txt'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')

    # Ensure the opener exist
    # import urllib.request as urllib_x
    from urllib.error import URLError  # NOQA
    # if urllib_x._opener is None:
    #     urllib_x.install_opener(urllib_x.build_opener())

    dpath = ub.ensure_app_cache_dir('ubelt', 'tests')
    fname = basename(url)
    fpath = join(dpath, fname)

    ub.delete(fpath)
    assert not exists(fpath)

    with pytest.raises(URLError):
        ub.download(url, fpath=fpath, verbose=1, timeout=1.0)


@pytest.mark.timeout(5)
def test_grabdata_fname_only():
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    # fname = 'mario.png'

    url = _demo_url()

    dpath = ub.ensure_app_cache_dir('ubelt')
    fname = 'custom_text.txt'
    fpath = join(dpath, fname)

    got_fpath = ub.grabdata(url, fname=fname)
    assert got_fpath == fpath
    assert exists(fpath)


@pytest.mark.timeout(5)
def test_grabdata_dpath_only():
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.ensure_app_cache_dir('ubelt', 'test')
    fname = basename(url)
    fpath = join(dpath, fname)

    got_fpath = ub.grabdata(url, dpath=dpath)
    assert got_fpath == fpath
    assert exists(fpath)


@pytest.mark.timeout(5)
def test_grabdata_fpath_and_dpath():
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    with pytest.raises(ValueError):
        ub.grabdata(url, fpath='foo', dpath='bar')


# @pytest.mark.timeout(5)
def test_grabdata_hash_typo():
    """
    CommandLine:
        xdoctest ~/code/ubelt/ubelt/tests/test_download.py test_grabdata_hash_typo --network

    """
    # url = 'https://www.dropbox.com/s/jl506apezj42zjz/ibeis-win32-setup-ymd_hm-2015-08-01_16-28.exe?dl=1'
    import hashlib
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')

    url = _demo_url()

    dpath = ub.ensure_app_cache_dir('ubelt')
    fname = basename(url)
    fpath = join(dpath, fname)

    for verbose in [0]:
        ub.delete(fpath)
        ub.delete(fpath + '.md5.hash')
        assert not exists(fpath)

        print('[STEP1] Downloading file, but we have a typo in the hash')
        with pytest.raises(RuntimeError):
            got_fpath = ub.grabdata(
                url, hash_prefix='545e3a51404f-typo-4e46aa65a70948e126',
                hasher=hashlib.md5(), verbose=verbose)
        assert exists(fpath)

        print('[STEP2] Fixing the typo recomputes the hash, but does not redownload the file')
        got_fpath = ub.grabdata(url,
                                hash_prefix='22d42eb002cefa81e9ad604ea57bc01d',
                                hasher=hashlib.md5(), verbose=verbose)
        assert got_fpath == fpath
        assert exists(fpath)

        # If we delete the .hash file we will simply recompute
        ub.delete(fpath + '.md5.hash')
        print('[STEP3] Deleting the hash file recomputes the hash')
        got_fpath = ub.grabdata(url, fpath=fpath,
                                hash_prefix='22d42eb002cefa81e9ad604ea57bc01d',
                                hasher=hashlib.md5(), verbose=verbose)
        assert exists(fpath + '.md5.hash')


class SingletonTestServer(ub.NiceRepr):
    """
    A singleton class used for testing.

    This could be done via a pytest fixture, but... I don't want to use
    fixtures until its easy and clear how to make an instance of them in an
    independent IPython session.

    CommandLine:
        xdoctest -m /home/joncrall/code/ubelt/tests/test_download.py SingletonTestServer

    Note:
        We rely on python process close mechanisms to clean this server up.
        Might need to re-investigate this in the future.

    Ignore:
        >>> self = SingletonTestServer.instance()
        >>> print('self = {!r}'.format(self))
        >>> url = self.urls[0]
        >>> print('url = {!r}'.format(url))
        >>> dl_file = ub.download(url)
    """

    _instance = None

    @classmethod
    def instance(cls):
        if cls._instance is not None:
            self = cls._instance
        else:
            self = cls()
            cls._instance = self
        return self

    def __nice__(self):
        return '{} - {}'.format(self.root_url, self.proc.returncode)

    def __init__(self):
        import requests
        import time
        import sys
        import ubelt as ub
        import socket
        from contextlib import closing
        def find_free_port():
            """
            References:
                https://stackoverflow.com/questions/1365265/on-localhost-how-do-i-pick-a-free-port-number
            """
            with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
                s.bind(('', 0))
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                return s.getsockname()[1]

        # Find an open port
        port = find_free_port()
        print('port = {!r}'.format(port))

        dpath = ub.ensure_app_cache_dir('ubelt/simple_server')

        if sys.version_info[0] == 2:
            server_cmd = [
                'python', '-m', 'SimpleHTTPServer', str(port)
            ]
        else:
            server_cmd = [
                'python', '-m', 'http.server', str(port)
            ]
        info = ub.cmd(server_cmd, detach=True, cwd=dpath)
        proc = info['proc']
        self.proc = proc
        self.dpath = dpath
        self.root_url = 'http://localhost:{}'.format(port)

        import platform
        is_pypy = platform.python_implementation() == 'PyPy'
        is_win32 = sys.platform.startswith('win32')

        if is_pypy and is_win32:
            # not sure why
            import pytest
            pytest.skip('not sure why download tests are failing on pypy win32')
            init_sleeptime = 0.5
            fail_sleeptime = 0.3
            timeout = 10
        else:
            init_sleeptime = 0.002
            fail_sleeptime = 0.01
            timeout = 1

        time.sleep(init_sleeptime)
        # Wait for the server to be alive
        status_code = None
        max_tries = 100
        for _ in range(max_tries):
            try:
                resp = requests.get(self.root_url, timeout=timeout)
            except requests.exceptions.ConnectionError:
                time.sleep(fail_sleeptime)
            else:
                status_code = resp.status_code
            if status_code == 200:
                break

        poll_ret = self.proc.poll()

        if poll_ret is not None:
            print('poll_ret = {!r}'.format(poll_ret))
            print(self.proc.communicate())
            raise AssertionError('Simple server did not start {}'.format(poll_ret))

        self.urls = []
        self.write_file()

    def write_file(self, filebytes=10, num_files=1):
        fnames = ['file_{}_{}.txt'.format(filebytes, i) for i in range(num_files)]
        for fname in fnames:
            # data = ''.join(random.choices(string.ascii_letters, k=filebytes))
            data = 'a' * filebytes
            fpath = join(self.dpath, fname)
            with open(fpath, 'w') as file:
                file.write(data)
        urls = ['{}/{}'.format(self.root_url, fname) for fname in fnames]
        self.urls.extend(urls)
        return urls


def test_local_download():
    server = SingletonTestServer.instance()
    url = server.write_file(filebytes=int(10 * 2 ** 20))[0]
    # also test with a timeout for lazy coverage
    ub.download(url, timeout=1000)


def _demo_url(num_bytes=None):
    REAL_URL = False
    if REAL_URL:
        url = 'http://i.imgur.com/rqwaDag.png'
        if not ub.argflag('--network'):
            pytest.skip('not running network tests')
    else:
        if num_bytes is None:
            url = SingletonTestServer.instance().urls[0]
        else:
            url = SingletonTestServer.instance().write_file(num_bytes)[0]
    return url


@pytest.mark.timeout(5)
def test_download_with_progkw():
    """
    Test that progkw is properly passed through to ub.download
    """
    url = _demo_url(128 * 10)
    dpath = ub.ensure_app_cache_dir('ubelt', 'tests')
    fname = basename(url)
    fpath = join(dpath, fname)
    with ub.CaptureStdout() as cap:
        ub.download(url, fpath=fpath, progkw={'verbose': 3, 'freq': 1, 'adjust': False}, chunksize=128)
    assert len(cap.text.split('\n')) > 10


def test_grabdata():
    # xdoctest: +REQUIRES(--network)
    import ubelt as ub
    # fname = 'foo.bar'
    # url = 'http://i.imgur.com/rqwaDag.png'
    # prefix1 = '944389a39dfb8fa9'
    fname = 'foo2.bar'
    url = _demo_url(128 * 11)
    prefix1 = 'b7fa848cd088ae842a89ef'
    fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1)
    stamp_fpath = fpath + '.sha512.hash'
    assert ub.readfrom(stamp_fpath) == prefix1
    # Check that the download doesn't happen again
    fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1)
    # todo: check file timestamps have not changed
    #
    # Check redo works with hash
    fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1, redo=True)
    # todo: check file timestamps have changed
    #
    # Check that a redownload occurs when the stamp is changed
    with open(stamp_fpath, 'w') as file:
        file.write('corrupt-stamp')
    fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1)
    assert ub.readfrom(stamp_fpath) == prefix1
    #
    # Check that a redownload occurs when the stamp is removed
    ub.delete(stamp_fpath)
    with open(fpath, 'w') as file:
        file.write('corrupt-data')
    assert not ub.hash_file(fpath, base='hex', hasher='sha512').startswith(prefix1)
    fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1)
    assert ub.hash_file(fpath, base='hex', hasher='sha512').startswith(prefix1)
    #
    # Check that requesting new data causes redownload
    #url2 = 'https://data.kitware.com/api/v1/item/5b4039308d777f2e6225994c/download'
    #prefix2 = 'c98a46cb31205cf'  # hack SSL
    # url2 = 'http://i.imgur.com/rqwaDag.png'
    # prefix2 = '944389a39dfb8fa9'
    url2, prefix2 = url, prefix1
    fpath = ub.grabdata(url2, fname=fname, hash_prefix=prefix2)
    assert ub.readfrom(stamp_fpath) == prefix2


def test_grabdata_delete_hash_stamp():
    import ubelt as ub
    fname = 'foo3.bar'
    url = _demo_url(128 * 12)
    prefix1 = '43f92597d7eb08b57c88b636'
    fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1)
    stamp_fpath = fpath + '.sha512.hash'
    ub.delete(stamp_fpath)
    fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1)


def test_download_with_io():
    import ubelt as ub
    import io
    url = _demo_url(128 * 3)
    file = io.BytesIO()
    fpath = ub.download(url, file)
    assert fpath is file
    file.seek(0)
    data = file.read()
    hashstr = ub.hash_data(data, hasher='sha1')
    assert hashstr.startswith('45a5c851bf12d1')


def test_download_with_sha1_hasher():
    import ubelt as ub
    url = _demo_url(128 * 4)
    ub.download(url, hasher='sha1', hash_prefix='164557facb7392')


if __name__ == '__main__':
    """
    CommandLine:
        pytest ubelt/tests/test_download.py
    """
    import xdoctest
    xdoctest.doctest_module(__file__)
